@@ -5,6 +5,13 @@ Numpy specific extensions
 from __future__ import absolute_import, division, print_function, unicode_literals
 import numpy as np
 
+try:
+    # The math variant only exists in Python 3+ but is faster for scalars
+    # so try and use it
+    from math import isclose
+except Exception:
+    from numpy import isclose
+
 
 def boolmask(indices, shape=None):
     """
@@ -42,7 +49,7 @@ def boolmask(indices, shape=None):
     """
     indices = np.asanyarray(indices)
     if indices.dtype.kind not in {'i', 'u'}:
-        indices = indices.astype(np.int)
+        indices = indices.astype(int)
     if shape is None:
         shape = indices.max() + 1
     mask = np.zeros(shape, dtype=np.bool)
@@ -201,27 +208,13 @@ def atleast_nd(arr, n, front=False):
         [[[[0]]]]
         [[[[[1]]]]]
 
-    Ignore:
-        # Hmm, mine is actually faster
-        %timeit atleast_nd(arr, 3)
-        %timeit np.atleast_3d(arr)
+    Notes:
+        Extensive benchmarks are in
+        kwarray/dev/bench_atleast_nd.py
 
-    Benchmark:
-
-        import ubelt
-        N = 100
-
-        t1 = ubelt.Timerit(N, label='mine')
-        for timer in t1:
-            arr = np.empty((10, 10))
-            with timer:
-                atleast_nd(arr, 3)
-
-        t2 = ubelt.Timerit(N, label='baseline')
-        for timer in t2:
-            arr = np.empty((10, 10))
-            with timer:
-                np.atleast_3d(arr)
+        These demonstrate that this function is statistically faster than the
+        numpy variants, although the difference is small.  On average this
+        function takes 480ns versus numpy which takes 790ns.
     """
     arr_ = np.asanyarray(arr)
     ndims = len(arr_.shape)
@@ -258,7 +251,7 @@ def argmaxima(arr, num, axis=None, ordered=True):
 
     Example:
         >>> # Test cases with axis=None
-        >>> arr = (np.random.rand(100) * 100).astype(np.int)
+        >>> arr = (np.random.rand(100) * 100).astype(int)
         >>> for num in range(0, len(arr) + 1):
         >>>     idxs = argmaxima(arr, num)
         >>>     idxs2 = argmaxima(arr, num, ordered=False)
@@ -267,7 +260,7 @@ def argmaxima(arr, num, axis=None, ordered=True):
 
     Example:
         >>> # Test cases with axis
-        >>> arr = (np.random.rand(3, 5, 7) * 100).astype(np.int)
+        >>> arr = (np.random.rand(3, 5, 7) * 100).astype(int)
         >>> for axis in range(len(arr.shape)):
         >>>     for num in range(0, len(arr) + 1):
         >>>         idxs = argmaxima(arr, num, axis=axis)
@@ -283,7 +276,7 @@ def argmaxima(arr, num, axis=None, ordered=True):
     if axis is None:
         axis_size = arr.size
         if num == 0:
-            return np.empty(0, dtype=np.int)
+            return np.empty(0, dtype=int)
         elif num == 1:
             # very fast
             idxs = np.array([arr.argmax(axis=axis)])
@@ -308,7 +301,7 @@ def argmaxima(arr, num, axis=None, ordered=True):
         if num == 0:
             new_shape = list(arr.shape)
             new_shape[axis] = 0
-            return np.empty(new_shape, dtype=np.int)
+            return np.empty(new_shape, dtype=int)
         elif num == 1:
             newshape = list(arr.shape)
             newshape[axis] = 1
@@ -360,13 +353,14 @@ def argminima(arr, num, axis=None, ordered=True):
     Args:
         arr (ndarray): input array
         num (int): number of minimum indices to return
-        axis (int|None): must be None for now
+        axis (int|None): axis to find minima over.
+            If None this is equivalent to using arr.ravel().
         ordered (bool): if False, returns the minimum elements in an arbitrary
             order, otherwise they are in ascending order. (Setting this to
             false is a bit faster).
 
     Example:
-        >>> arr = (np.random.rand(100) * 100).astype(np.int)
+        >>> arr = (np.random.rand(100) * 100).astype(int)
         >>> for num in range(0, len(arr) + 1):
         >>>     idxs = argminima(arr, num)
         >>>     assert np.all(arr[idxs] == np.array(sorted(arr)[:len(idxs)])), 'ordered=True must return in order'
@@ -374,37 +368,173 @@ def argminima(arr, num, axis=None, ordered=True):
         >>>     assert sorted(idxs2) == sorted(idxs), 'ordered=False must return the right idxs, but in any order'
 
     Example:
-        >>> arr = (np.random.rand(32, 32) * 100).astype(np.int)
-        >>> argminima(arr, 10)
+        >>> # Test cases with axis
+        >>> from kwarray.util_numpy import *  # NOQA
+        >>> arr = (np.random.rand(3, 5, 7) * 100).astype(int)
+        >>> # make a unique array so we can check argmax consistency
+        >>> arr = np.arange(3 * 5 * 7)
+        >>> np.random.shuffle(arr)
+        >>> arr = arr.reshape(3, 5, 7)
+        >>> for axis in range(len(arr.shape)):
+        >>>     for num in range(0, len(arr) + 1):
+        >>>         idxs = argminima(arr, num, axis=axis)
+        >>>         idxs2 = argminima(arr, num, ordered=False, axis=axis)
+        >>>         print('idxs = {!r}'.format(idxs))
+        >>>         print('idxs2 = {!r}'.format(idxs2))
+        >>>         assert idxs.shape[axis] == num
+        >>>         assert idxs2.shape[axis] == num
+        >>>         # Check if argmin argrees with -argmax
+        >>>         idxs3 = argmaxima(-arr, num, axis=axis)
+        >>>         assert np.all(idxs3 == idxs)
+
+    Example:
+        >>> arr = np.arange(20).reshape(4, 5) % 6
+        >>> argminima(arr, axis=1, num=2, ordered=False)
+        >>> argminima(arr, axis=1, num=2, ordered=True)
+        >>> argmaxima(-arr, axis=1, num=2, ordered=True)
+        >>> argmaxima(-arr, axis=1, num=2, ordered=False)
     """
     if axis is not None:
-        raise NotImplementedError('axis must be None for now')
-    # Gets top N maximum or minimum indices
-    if num < 0:
-        raise IndexError
-    elif num == 0:
-        return np.empty(0, dtype=np.int)
-    elif num == 1:
-        # very fast
-        idxs = np.array([arr.argmin(axis=axis)])
-    elif num < len(arr):
-        # argpartition is almost what we want, and its faster than argsort
-        kth = num
-        part_idxs = np.argpartition(arr, kth=kth, axis=axis)
-        idxs = part_idxs[:kth]
-        if ordered:
-            sortx = arr.take(idxs, axis=axis).argsort()
-            idxs = idxs.take(sortx)
-    else:
-        # sort all the indices
-        if ordered:
-            idxs = arr.argsort(axis=axis)[0:num]
+        # TODO: implement this directly for argminima
+        # if 0:
+        #     return argmaxima(-arr, num=num, axis=axis, ordered=ordered)
+        # # raise NotImplementedError('axis must be None for now, can use argmaxima(-arr) with axis in the meantime')
+        # raise NotImplementedError('probably need to switch some cases')
+        axis_size = arr.shape[axis]
+        if num == 0:
+            new_shape = list(arr.shape)
+            new_shape[axis] = 0
+            return np.empty(new_shape, dtype=int)
+        elif num == 1:
+            newshape = list(arr.shape)
+            newshape[axis] = 1
+            idxs = arr.argmin(axis=axis).reshape(*newshape)
+        elif num < axis_size:
+            # TODO: is there a better implementation for this case?
+            kth = num
+            argmin_slice = slice(0, kth)
+            part_idxs = np.argpartition(arr, kth=kth, axis=axis)
+            fancy_index = [slice(None)] * (axis + 1)
+            fancy_index[axis] = argmin_slice
+            idxs = part_idxs[tuple(fancy_index)]
+            if ordered:
+                # move the axis of interest to the back
+                idxs_swap = idxs.swapaxes(-1, axis)
+                idxs2d = idxs_swap.reshape(-1, num)
+                arr2d = arr.swapaxes(-1, axis).reshape(-1, arr.shape[axis])
+                # now ensure each row is in order
+                new_idxs2d = []
+                for a, i in zip(arr2d, idxs2d):
+                    sortx = a[i].argsort()[::+1]
+                    new_idxs2d.append(i[sortx])
+                new_idxs2d = np.array(new_idxs2d)
+                # transform back to original shape
+                idxs = new_idxs2d.reshape(*idxs_swap.shape).swapaxes(-1, axis)
         else:
-            # Arbitrary order is allowed, so cheat
-            idxs = np.arange(arr.size)
+            # sort all the indices
+            if ordered:
+                idxs = arr.argsort(axis=axis)
+                fancy_index = [slice(None)] * (axis + 1)
+                fancy_index[axis] = slice(None, None, +1)
+                idxs = idxs[tuple(fancy_index)]
+            else:
+                # Arbitrary order is allowed, so cheat
+                idxs = np.arange(arr.shape[axis])
+                newshape = [1] * len(arr.shape)
+                newshape[axis] = arr.shape[axis]
+                repeats = list(arr.shape)
+                repeats[axis] = 1
+                idxs = np.tile(idxs.reshape(*newshape), repeats)
+    else:
+        # Gets top N maximum or minimum indices
+        if num < 0:
+            raise IndexError
+        elif num == 0:
+            return np.empty(0, dtype=int)
+        elif num == 1:
+            # very fast
+            idxs = np.array([arr.argmin(axis=axis)])
+        elif num < len(arr):
+            # argpartition is almost what we want, and its faster than argsort
+            kth = num
+            part_idxs = np.argpartition(arr, kth=kth, axis=axis)
+            idxs = part_idxs[:kth]
+            if ordered:
+                sortx = arr.take(idxs, axis=axis).argsort()
+                idxs = idxs.take(sortx)
+        else:
+            # sort all the indices
+            if ordered:
+                idxs = arr.argsort(axis=axis)[0:num]
+            else:
+                # Arbitrary order is allowed, so cheat
+                idxs = np.arange(arr.size)
     return idxs
 
 
+# def unique_axis(arr, axis):
+#     """
+#     Like unique and unique_rows, but looks over any single axis
+
+#     Args:
+#         axis (int): axis to consider as a whole unit
+
+#     References:
+#         https://stackoverflow.com/questions/16970982/find-unique-rows-in-numpy-array
+
+#     >>> import kwarray
+#     rng = kwarray.ensure_rng(0)
+#     arr = rng.randint(0, 2, size=(2, 2, 3, 2, 2)).T
+#     axis = 2
+#     """
+#     raise NotImplementedError
+#     orig_shape = arr.shape
+#     axis_len = arr.shape[axis]
+#     dtype_view = np.dtype((np.void, arr.dtype.itemsize * axis_len))
+#     arr_swap = arr.swapaxes(axis, 0)
+#     arr_swap_flat = arr_swap.reshape(axis_len, -1).T
+#     arr_swap_view = np.ascontiguousarray(arr_swap).view(dtype_view)
+#     arr_swap_view_unique = np.unique(arr_swap_view)
+#     arr_swap_unique = arr_swap_view_unique.view(arr.dtype)
+#     arr_unique =
+#     arr_swap_unique.reshape(-1, axis_len).T
+#     # return (
+#     #     .reshape(-1, arr.shape[1])
+#     # )
+
+def unique_rows(arr, ordered=False):
+    """
+    Like unique, but works on rows
+
+    Args:
+        arr (ndarray): must be a contiguous C style array
+        ordered (bool): if true, keeps relative ordering
+
+    References:
+        https://stackoverflow.com/questions/16970982/find-unique-rows-in-numpy-array
+
+    Example:
+        >>> import kwarray
+        >>> from kwarray.util_numpy import *  # NOQA
+        >>> rng = kwarray.ensure_rng(0)
+        >>> arr = rng.randint(0, 2, size=(12, 3))
+        >>> arr_unique = unique_rows(arr)
+        >>> print('arr_unique = {!r}'.format(arr_unique))
+    """
+    dtype_view = np.dtype((np.void, arr.dtype.itemsize * arr.shape[1]))
+    arr_view = arr.view(dtype_view)
+    if ordered:
+        arr_view_unique, idxs = np.unique(arr_view, return_index=True)
+        arr_flat_unique = arr_view_unique.view(arr.dtype)
+        arr_unique = arr_flat_unique.reshape(-1, arr.shape[1])
+        arr_unique = arr_unique[np.argsort(idxs)]
+    else:
+        arr_view_unique = np.unique(arr_view)
+        arr_flat_unique = arr_view_unique.view(arr.dtype)
+        arr_unique = arr_flat_unique.reshape(-1, arr.shape[1])
+    return arr_unique
+
+
 def arglexmax(keys, multi=False):
     """
     Find the index of the maximum element in a sequence of keys.
@@ -423,7 +553,7 @@ def arglexmax(keys, multi=False):
     Example:
         >>> k, N = 100, 100
         >>> rng = np.random.RandomState(0)
-        >>> keys = [(rng.rand(N) * N).astype(np.int) for _ in range(k)]
+        >>> keys = [(rng.rand(N) * N).astype(int) for _ in range(k)]
         >>> multi_idx = arglexmax(keys, multi=True)
         >>> idxs = np.lexsort(keys)
         >>> assert sorted(idxs[::-1][:len(multi_idx)]) == sorted(multi_idx)
@@ -432,7 +562,7 @@ def arglexmax(keys, multi=False):
         >>> import ubelt as ub
         >>> k, N = 100, 100
         >>> rng = np.random
-        >>> keys = [(rng.rand(N) * N).astype(np.int) for _ in range(k)]
+        >>> keys = [(rng.rand(N) * N).astype(int) for _ in range(k)]
         >>> for timer in ub.Timerit(100, bestof=10, label='arglexmax'):
         >>>     with timer:
         >>>         arglexmax(keys)
@@ -493,8 +623,8 @@ def normalize(arr, mode='linear', alpha=None, beta=None, out=None):
 
         >>> raw_f = np.random.rand(8, 8) * 100
         >>> norm_f = normalize(raw_f)
-        >>> assert np.isclose(norm_f.min(), 0)
-        >>> assert np.isclose(norm_f.max(), 1)
+        >>> assert isclose(norm_f.min(), 0)
+        >>> assert isclose(norm_f.max(), 1)
 
         >>> raw_u = (np.random.rand(8, 8) * 255).astype(np.uint8)
         >>> norm_u = normalize(raw_u)
@@ -620,6 +750,10 @@ def normalize(arr, mode='linear', alpha=None, beta=None, out=None):
             # This chooses alpha such the original min/max value will be pushed
             # towards -1 / +1.
             alpha = max(abs(old_min - beta), abs(old_max - beta)) / 6.212606
+
+        if isclose(alpha, 0):
+            alpha = 1
+
         energy = float_out
         energy -= beta
         energy /= alpha
@@ -638,6 +772,7 @@ def normalize(arr, mode='linear', alpha=None, beta=None, out=None):
 
     if float_out is not out:
         out[:] = float_out.astype(out.dtype)
+
     return out
 
 
